Require Import Vbase Relations ClassicalDescription ClassicalChoice.
Require Import List Permutation Sorting extralib. 
Require Import c11 exps.
Require Import GpsSepAlg GpsModel GpsModelLemmas GpsLogic.

Set Implicit Arguments.

(* We can construct a snapshot monoid from any partially ordered set P
   with a bottom and the additional property that for all x,y,z in P, if x <= z and y <= z
   then x <= y or y <= x. That is, if two elements share a common upper bound,
   then they are comparable.
 *)

Record cbcposet_mod := mk_cbcposet
  { cbcposet :> Type
  ; cbcposet_bot: cbcposet
  ; cbcposet_lt : cbcposet -> cbcposet -> Prop
  ; cbcposet_bot_lt: forall x, cbcposet_lt cbcposet_bot x
  ; cbcposet_refl : forall x, cbcposet_lt x x
  ; cbcposet_trans: forall x y z, cbcposet_lt x y -> cbcposet_lt y z -> cbcposet_lt x z
  ; cbcposet_antisym: forall x y, cbcposet_lt x y -> cbcposet_lt y x -> x = y
  ; cbcposet_cbc: forall x y z, cbcposet_lt x z -> cbcposet_lt y z -> (cbcposet_lt x y \/ cbcposet_lt y x)
  }.

Inductive view {A: Type} : Type :=
  | snapshot: A -> view
  | master: A -> view.

Section view_plus_properties.
  
Variable p: cbcposet_mod.
Definition view_emp := snapshot (cbcposet_bot p).
Definition view_strip (x: @view (cbcposet p)) := 
  match x with 
      | snapshot a => snapshot a
      | master a => snapshot a
  end.

Definition view_plus (x y: @view (cbcposet p)) : option (@view (cbcposet p)) :=
    match x, y with
        | master _, master _ => None
        | snapshot a, master b
        | master b, snapshot a =>
          if excluded_middle_informative (cbcposet_lt p a b) then
            Some (master b)
          else
            None
        | snapshot a, snapshot b =>
          if excluded_middle_informative (cbcposet_lt p a b) then
            Some (snapshot b)
          else
            if excluded_middle_informative (cbcposet_lt p b a) then
              Some (snapshot a)
            else
              None
    end.

Lemma view_plusC: forall x y, view_plus x y = view_plus y x.
Proof.
  destruct x, y; simpl; auto. desf. repeat apply f_equal.
  eapply cbcposet_antisym; auto.
Qed.

Lemma view_plusI: forall x, view_plus view_emp x = Some x.
Proof.
  destruct x; simpl; auto.
  desf. repeat apply f_equal. 
  eapply cbcposet_antisym; auto. eapply cbcposet_bot_lt.
  exfalso; eapply n. eapply cbcposet_bot_lt.
  desf. exfalso; eapply n. eapply cbcposet_bot_lt.
Qed.
  
Lemma view_plus_invSS: 
  forall a b v, view_plus (snapshot a) (snapshot b) = Some v 
                -> (v = snapshot a /\ cbcposet_lt p b a) \/ (v = snapshot b /\ cbcposet_lt p a b).
  ins; desf; auto.
Qed.

Lemma view_plus_invMS: 
  forall a b v, view_plus (master a) (snapshot b) = Some v 
                -> (v = master a /\ cbcposet_lt p b a).
  ins; desf; auto.
Qed.

Lemma view_plus_invMS_none: 
  forall a b, view_plus (master a) (snapshot b) = None 
                -> (~ cbcposet_lt p b a).
  ins; desf; auto.
Qed.
  
Lemma view_plus_invSM: 
  forall a b v, view_plus (snapshot b) (master a) = Some v 
                -> (v = master a /\ cbcposet_lt p b a).
  ins; desf; auto.
Qed.
 
Lemma view_plusA : 
      forall x y xy (P1: view_plus x y = Some xy) z xyz 
                    (P2: view_plus xy z = Some xyz),
    exists yz, view_plus x yz = Some xyz /\ 
               view_plus y z = Some yz.
Proof.
  destruct x, y, z; ins; desf; ins; desf;
  try (eexists; split; [|done]; simpl; desf);
  exfalso; try (by destruct p; ins; eauto).
  by edestruct (cbcposet_cbc p c0 c1); eauto.
  by edestruct (cbcposet_cbc p c0 c1); eauto.
Qed.

Lemma view_plusA2 : 
      forall x y xy (P1: view_plus x y = Some xy) z 
                    (P2: view_plus xy z = None),
        view_plus y z = None
        \/ exists yz, view_plus x yz = None /\ 
                      view_plus y z = Some yz.
Proof.
  destruct x, y; intros; simpl in P1; desf; eauto.
  * ins; desf; eauto. 
    by right; exists (snapshot c2); desf.
    by destruct n2, p; ins; eauto.
    by right; exists (master c2); desf.
  * ins; desf; eauto.
    by right; exists (snapshot c2); desf.
    by destruct n0, p; ins; eauto.
    by right; exists (master c2); desf.
Qed.

Lemma view_eq_empD : 
      forall g g', view_plus g g' = Some view_emp -> 
                   g = view_emp.
Proof.
  destruct g, g'; simpl; desf.
  intros. inversion H. subst. eapply f_equal. eapply cbcposet_antisym.
  eauto. eapply cbcposet_bot_lt.
  intros. desf.
Qed.
  
Lemma view_strip_emp :
      view_strip view_emp = view_emp.
Proof.                              
  auto.
Qed.

Lemma view_strip_plus :
      forall x y xy, 
        view_plus x y = Some xy ->
        view_plus (view_strip x) (view_strip y) = Some (view_strip xy).
Proof.
  intros; 
  destruct x, y; simpl in *. 
  desf. 
  desf.
  desf. simpl. repeat apply f_equal. eapply (cbcposet_antisym p); eauto.
  desf.
Qed.

Lemma view_plus_strip_idemp :
      forall x, view_plus x (view_strip x) = Some x.
Proof.
  intros; destruct x; simpl in *; desf; auto.
  exfalso; eapply n; eapply (cbcposet_refl). 
  exfalso; eapply n; eapply (cbcposet_refl). 
Qed.
 
Lemma view_strip_max :
      forall x y xy z, 
        view_plus x y = Some xy ->
        view_plus xy z = Some x ->
        view_strip y = y.
Proof.
  intros; destruct x, y. auto.
  simpl in H. desf. destruct z; simpl in *; desf.  auto.
  simpl in H. desf.
Qed.

Lemma view_pcancel :
      forall g g1 g2 gg,
        view_plus g g1 = Some gg ->
        view_plus g g2 = Some gg -> 
        view_plus (view_strip g) g1 = 
        view_plus (view_strip g) g2.
Proof.
  destruct g.
  intros. unfold view_strip. rewrite H. rewrite H0. auto.
  intros. unfold view_strip.
  destruct g1. destruct g2.
  exploit (view_plus_invMS); eauto; intros. clear H0.
  exploit (view_plus_invMS); eauto; intros. clear H.
  desf.
  simpl. desf.
  repeat apply f_equal. 
  exploit cbcposet_antisym. eapply x3. eapply c3. 
  exploit cbcposet_antisym. eapply x2. eapply c2. 
  intros. subst. auto.
  exploit cbcposet_antisym. eapply x2. eapply c2. 
  intros. subst. auto.
  exploit cbcposet_antisym. eapply x3. eapply c3. 
  intros. subst. auto.
  desf. desf.
Qed.

End view_plus_properties.

Definition sa_view (p: cbcposet_mod) :=
  {| salg := @view (cbcposet p)
   ; salg_emp := view_emp p
   ; salg_plus := view_plus p
   ; salg_strip := view_strip p 
   ; salg_plusC := view_plusC p
   ; salg_plusI := view_plusI p
   ; salg_plusA := view_plusA p
   ; salg_plusA2 := view_plusA2 p
   ; salg_eq_empD := view_eq_empD (p:=p) 
   ; salg_strip_emp := view_strip_emp p
   ; salg_strip_plus := view_strip_plus p
   ; salg_plus_strip_idemp := view_plus_strip_idemp p
   ; salg_strip_max := view_strip_max p
   ; salg_pcancel := view_pcancel p
  |}.

Definition snapshotG {p} (H: cbcposet p) := 
  Some (existT [eta salg] (sa_view p) (snapshot H)).
Definition masterG {p} (H: cbcposet p) := 
  Some (existT [eta salg] (sa_view p) (master H)).
Definition Snapshot {p} a (H: cbcposet p) := Rgres (gres_one a (snapshotG H)).
Definition Master {p} a (H: cbcposet p) := Rgres (gres_one a (masterG H)).
